(* Parsing the command-line *)

let rank = ref 3
let only = ref 0
let digits = ref ""

let arg_spec = [
  "-rank", Arg.Set_int rank, "the rank of the grid 3(default),4,5";
  "-only", Arg.Set_int only, "stops after n solutions";
  "-digits", Arg.Set_string digits, "the string of digits";
]

let usage_msg = "Af-sudoku\n"

let usage _ = Arg.usage arg_spec usage_msg; exit 1

let () = 
  Arg.parse arg_spec usage usage_msg;
  (match !rank with 3|4|5 -> ()
    | _ -> Printf.eprintf "Error: rank must be 3,4 or 5\n"; exit 2);
  (if !only < 0 then
    (Printf.eprintf "Error: only myst be >= 0\n"; exit 2));
  (if !digits <> "" && String.length !digits != !rank * !rank then
     (Printf.eprintf "Error: wrong digit string (length=%i)\n" (String.length !digits); exit 2))

    

let rank = !rank

(* Global constants *)

let chars = 
  if !digits <> "" then !digits
  else
  match rank with
    | 3 -> "123456789"
    | 4 -> "123456789ABCDEFG"
    | 5 -> "ABCDEFGHIJKLMNOPQRSTUVWXY"
    | _ -> assert false

let rank2 = rank * rank
let rank4 = rank2 * rank2

(* Matrix.

   A grid is represented as a one dimensional integer array of size rank4+1.
   The cell 0..rank4-1 are bit fields which indicate the possible
   digits for the corresponding cell. The last cell of the grid
   store the number of resolved cells. *)

let coord x y = x + rank2 * y

(* Bit-field manipulation *)

(* rank of the least significant 1 bit *)
let bit_rank k =
  let rec aux i k = if (k = 1) then i else aux (succ i) (k lsr 1) in
  aux 0 k

(* keeps only the least significant 1 bit *)
let least_bit n = n land (lnot (n - 1))

(* is there exactly one 1 bit ? *)
let only_one_1 n = (n != 0) && (n == least_bit n)

(* number of 1 bits *)

let count_bits n =
  let rec l i n =if n = 0 then i else l (succ i) (n land lnot (least_bit n)) in
  l 0 n

let nb_bits_byte = Array.init 256 count_bits

let nb_bits n =
  nb_bits_byte.(n land 0xff) +
  nb_bits_byte.((n lsr 8) land 0xff) +
  nb_bits_byte.((n lsr 16) land 0xff) +
  nb_bits_byte.((n lsr 24) land 0xff)


(* Prints a grid *)

let print a =
  for y = 0 to rank2 - 1 do
    for x = 0 to rank2 - 1 do
      let i = coord x y in
      if only_one_1 a.(i) then print_char chars.[bit_rank a.(i)]
      else print_char '.'
    done;
    print_newline ();
  done

let solution a =
  print a; print_endline "---";
  if !only > 0 then (decr only; if !only == 0 then exit 0)


exception Impossible

(* This function is called when one cell has been resolved. It propagates
   the constraints. *)

let rec propagate a i =
  a.(rank4) <- a.(rank4) + 1;
  let x = i mod rank2 and y = i / rank2 in
  let m = lnot a.(i) in
  let f x' y' = 
    let i' = coord x' y' in
    if i != i' then
      let old = a.(i') in
      let n = old land m in
      if n = 0 then raise Impossible
      else if n != old 
      then (a.(i') <- n; if only_one_1 n then propagate a i')
  in
  for y' = 0 to rank2 - 1 do f x y' done;
  for x' = 0 to rank2 - 1 do f x' y done;
  let x0 = (x / rank) * rank and y0 = (y / rank) * rank in
  for x' = x0 to x0+(rank-1) do for y' = y0 to y0+(rank-1) do f x' y' done done
    
let restrict a i n = a.(i) <- n; if only_one_1 n then propagate a i
let resolve a i n = a.(i) <- n; propagate a i

(* A block is a group of rank2 cells (a lines, a column, a box) such
   that each digit must appear exactly once in this group of
   cells. The function sets digits according to this rule and returns
   the coordinates of the modified cells. *)
    
let find_in_block a0 block =
  let change = ref false in
  let rec check_digit k =
    if (k = rank2) then ()
    else none_found (1 lsl k) k 0
  and none_found m k i =
    if (i = rank2) then raise Impossible
    else let n = a0.(block.(i)) in
    if (n == m) then check_digit (succ k)
    else if (n land m != 0) then found_one m k i (succ i)
    else none_found m k (succ i)
  and found_one m k i0 i =
    if (i = rank2) then 
      (resolve a0 block.(i0) m; change := true; check_digit (succ k))
    else if (a0.(block.(i)) land m != 0) then check_digit (succ k)
    else found_one m k i0 (succ i)
  in
  check_digit 0;
  !change


let blocks = Array.concat [
  Array.init rank2 
    (fun y -> Array.init rank2 (fun x -> coord x y)); (* lines *)
  Array.init rank2 
    (fun x -> Array.init rank2 (fun y -> coord x y)); (* columns *)
  Array.init rank2 
    (fun i ->
       let x0 = (i mod rank) * rank
       and y0 = (i / rank) * rank in
       Array.init rank2 (fun j -> coord (x0 + (j mod rank)) (y0 + (j / rank)))
    )  (* boxes *)
]

let find_in_blocks a =
  let f = ref true in
  while !f do
    f := false;
    for i = 0 to Array.length blocks - 1 do
      if find_in_block a blocks.(i) then f := true;
    done;
  done


(* Chains.

   For each unresolved cell, try all the possible choices one by one
   and deduce all the consequences from each one.

   - If we get a contradiction, we can reduce the number of possible
   choices.

   - For all the choices which does not imply an immediate contradiction,
   we consider the set of remaining choices for all cells, and we
   take the unions of all these choices. This set might be smaller
   than the original set of choices.

   - We also compute a best branching cell. For each cell,
   we consider the minimal number of cells fixed as a consequence
   of fixing the value for this cell. The best cell is the one which
   maximize this number.

   - We iterate as long as something changed.
*)

let rec chains a =
  let change = ref false in
  let max_min_fixed = ref a.(rank4) in
  let best = ref (-1) in
  for i = 0 to rank4 - 1 do match nb_bits a.(i) with 
    | 0 -> raise Impossible
    | 1 -> ()
    | n ->
 	let possible = Array.create rank4 0 in
	let min_fixed = ref max_int in
	for k = 0 to rank2 - 1 do
	  let m = 1 lsl k in
	  if a.(i) land m <> 0 then
	    try 
	      let a' = Array.copy a in
	      resolve a' i m;
	      find_in_blocks a';
	      for z = 0 to rank2 - 1 do
		possible.(z) <- possible.(z) lor a'.(z)
	      done; 
	      if a'.(rank4) < !min_fixed then min_fixed := a'.(rank4);
	    with Impossible ->
	      let n = a.(i) land lnot m in
	      restrict a i n;
	      change := true
	done;
	if !min_fixed > !max_min_fixed then
	  (max_min_fixed := !min_fixed;
	   best := i);
	for z = 0 to rank2 - 1 do
	  let o = a.(z) in
	  let n = possible.(z) land o in
	  if n != o then (restrict a z n; change := true);
	done;  
  done;
  if !change then (find_in_blocks a; chains a) else !best


(* Find a cell whose number of remaining choices is minimal and >= 2 *)

let least_var a =
  let cur_min = ref max_int and cur = ref (-1) in
  (try for i = 0 to rank4 - 1 do
     let n = nb_bits a.(i) in
     if n = 0 then raise Impossible;
     if (n = 2) then (cur := i; cur_min := 2; raise Exit);
     if (n > 1) && (n < !cur_min) then (cur_min := n; cur := i)
   done with Exit -> ());
  !cur

(* Find a cell such that the minimum number of cells fixed as a consequence 
   of fixing it is maximal. *)

let rec main_loop a =
  find_in_blocks a;
  let i = chains a in
  if (i < 0) then solution a
  else
    for k = 0 to rank2 - 1 do
      let m = 1 lsl k in
      if a.(i) land m <> 0 then
	try 
	  let a' = 
	    if a.(i) land lnot (m lsl 1 - 1) = 0 then a else Array.copy a in
	  resolve a' i m;
	  main_loop a'
	with Impossible -> ()
    done
  
let () =
  let a = Array.create (succ rank4) ((1 lsl rank2) - 1) in
  a.(rank4) <- 0;
  let placed = ref [] in
  let rec loop i = if i == rank4 then () else  match input_char stdin with
    | '0' | '.' -> loop (succ i)
    | '\n' | '\r' ->
	if (i mod rank2 != 0) then
	  (Printf.eprintf "End-of-line on the middle of a line"; exit 2);
	loop i
    | c ->
	let n = try String.index chars c with Not_found -> -1 in
	if n >= 0 then (a.(i) <- 1 lsl n; placed := i :: !placed;
			loop (succ i))
	else loop i
  in
  loop 0;
  (try List.iter (propagate a) !placed;
   with Impossible -> exit 0);
  main_loop a
